/* $Id: inout.c,v 1.12 2008/03/30 23:01:53 pnykanen Exp $ */

#include "inout.h"

char *fileExtension = ".ph\0\0";

/* for getting current path, used in file reading */
void getPath(char *pathname)
{
  if (getcwd(pathname, sizeof(char)*MAXPATHLEN) == NULL )
    {
      printf("Error while trying to get current path!\n");
      exit(-1);
    }
}

phoneme** readDatasetFile(char *filename, int *count)
{
  char *tmp = NULL;
  double *vec;
  double *frame;
  int i, j, lines;
  int dim = 0, inc = 0, err = 0;
  FILE *fp = fopen(filename,"r");
  phoneme **phonemes = NULL;

  if (fp == NULL)
    {
      printf("Error while trying to open file %s.\n", filename);
      exit(-1);
    }

  lines = countLines(filename);

  if ((phonemes = newPhonemes(lines)) == NULL)
    {
      printf("Error while trying to allocate memory for phoneme pointers.\n");
      exit(2);
    }

  tmp = (char*)malloc(20*sizeof(char));
  vec = (double*)malloc(dim*sizeof(double));
  frame = (double*)malloc(sizeof(double));

  i = 0;
  while (vec != NULL)
    {
      vec = ReadVector(fp, &dim, &inc, &err);
      
      if (vec == NULL)
	break;

      sprintf(tmp, "ts-line %d", i);

      phonemes[i] = newPhoneme(tmp, 1);
      phonemes[i]->phNo = i;

      for (j=0; j < 20; j++)
	tmp[j] = '\0';

      for (j=0; j < dim; j++)
	{
	  frame[0] = vec[j];
	  insertFrame(phonemes[i], frame);
	}
      i++;
    }

  fclose(fp);

  free(tmp);
  free(vec);
  free(frame);

  *count = i;

  return phonemes;
} 

/* for reading single datafile to memory */
phoneme* readFile(char *filename, int dim)
{
  double *vec;
  int decided_dim = dim, inc = 0, err = 0;
  FILE *fp = fopen(filename,"r");
  phoneme *ptr = newPhoneme(filename, dim);

  if ( (fp == NULL) || (ptr == NULL) )
    {
      printf("Error while trying to read file %s\n", filename);
      return NULL;
    }

  vec = (double*)malloc(dim*sizeof(double));
  while (vec != NULL)
    {
      vec = ReadVector(fp, &dim, &inc, &err);
      if (vec == NULL)
	break;

      /* ONLY INSERTS LINE WITH CORRECT AMOUNT OF VALUES */
      if ( (dim == decided_dim) && (inc == 0) && (err == 0) )
	insertFrame(ptr, vec);
    }

  fclose(fp);
  return ptr;
}

/* for reading all datafiles in given directory to memory */
phoneme** readAllFiles(char *pathname, int *ph_sum, int *maxPathSize)
{
  int count;
  int dim;
  int i = 0;
  int file_select();
  phoneme **phonemes = NULL;
  struct direct **files;

  count = scandir(pathname, &files, file_select, alphasort);

  if (count <= 0)
    {
      printf("No files in this directory.\n");
      exit(1);
    }

  //printf("%d files found!\n", count);
      
  if ((phonemes = newPhonemes(count)) == NULL)
    {
      printf("Error while trying to allocate memory for phoneme pointers.\n");
      exit(2);
    }

  dim = decideDimensions(files[0]->d_name);
  if (dim < 1)
    {
      printf("Error while trying to decide dimensions of files.\n");
      exit(3);
    }

  for (i = 0; i < count; i++)
    {
      if ((phonemes[*ph_sum] = readFile(files[i]->d_name, dim)) == NULL)
	{
	  printf("Error while trying to read file %s.\n", files[i]->d_name);
	  exit(3);
	}

      if (phonemes[*ph_sum]->frameAmount == 0)
	{
	  printf("\nFile %s was empty or did not have matching dimensions, skipping!\n", 
		 files[i]->d_name);
	  free(files[i]);
	  files[i] = NULL;
	  continue;
	}
	
      phonemes[*ph_sum]->phNo = *ph_sum;
      (*ph_sum)++;

      free(files[i]);
      files[i] = NULL;
    }
  //printf(" done.\n");
  
  free(files);
  files = NULL;

  *maxPathSize = 2*countMaxFrameSize(phonemes, count);

  return phonemes;
}

int decideDimensions(char *filename)
{
  int idx, table_size = 50;
  int **dim_table, *larger;
  double *vec;
  int i = 0, max = 0;
  int dim = 0, inc = 0, err = 0;
  FILE *fp = fopen(filename,"r");

  dim_table = (int**)malloc(2*sizeof(int*));
  dim_table[0] = (int*)calloc(table_size,sizeof(int));
  dim_table[1] = (int*)calloc(table_size,sizeof(int));
  vec = (double*)malloc(dim*sizeof(double));
    
  if ( (fp == NULL) 
       || (dim_table == NULL)
       || (dim_table[0] == NULL)
       || (dim_table[1] == NULL)
       || (vec == NULL) )
    {
      printf("Error while trying to decide file dimensions from %s\n", filename);
      exit(1);
    }

  while (vec != NULL)
    {
      vec = ReadVector(fp, &dim, &inc, &err);
      if (vec == NULL)
	break;

      if ((idx = inList(dim_table[0], table_size, dim)) >= 0)
	{
	  dim_table[1][idx] += 1;
	}
      else 
	{
	  i=0;
	  while (dim_table[0][i] != 0)
	    {
	      if (i != table_size) 
		{
		  i++;
		}
	      else
		{
		  larger = (int*)calloc(table_size*2, sizeof(int));
		  memcpy(larger, dim_table[0], table_size * sizeof(int));
		  free(dim_table[0]);
		  dim_table[0] = larger;
		  larger = (int*)calloc(table_size*2, sizeof(int));
		  memcpy(larger, dim_table[1], table_size * sizeof(int));
		  free(dim_table[1]);
		  dim_table[1] = larger;
		  table_size *= 2;
		}
	    }
	  dim_table[0][i] = dim;
	  dim_table[1][i] += 1;
	}
    }
  fclose(fp);
  free(vec);

  if (dim_table[0][0] == 0)
    {
      printf("Error while trying to decide file dimensions from %s\n, empty file?", filename);
      exit(2);
    }

  i=0; idx = 0; max = 0;
  while (dim_table[0][i] != 0)
    {
      if (dim_table[1][i] > max)
	{
	  max = dim_table[1][i];
	  idx = i;
	}
      i++;
    }

  max = dim_table[0][idx];
  //printf("Calculated dimensions from %s: %d\n", filename, max);

  free(dim_table[0]);
  free(dim_table[1]);
  free(dim_table);

  return max;
}

/* for selecting wanted files, for example files ending in ".ph" */
int file_select(struct direct *entry)
{
  char *ptr;
  //  char *rindex(char *s, int c);

  if ( (strcmp(entry->d_name,".") == 0) || (strcmp(entry->d_name,"..") == 0) )
    return FALSE;

  /* Check for filename extensions */
  //  ptr = rindex(entry->d_name, (int)'.'); /* TODO: USE EXTENSION GIVEN BY USER */
  if ( (ptr != NULL)) //&& (strcmp(ptr,fileExtension) == 0) )
    return TRUE;
  else
    return FALSE;
}

/* for printing single frame */
void printFrame(double *fp, int w)
{
  unsigned short i;
  for(i=0; i < w; i++)
    printf("%.3f ", fp[i]);
  printf("\n");
}

/* for printing all frames of phoneme */
void printFrames(phoneme *p)
{
  unsigned short i;
  for(i=0; i < p->frameAmount; i++)
    printFrame(p->frames[i], p->dimensions);
  printf("\n");
}

void printPath(char *array)
{
  int i;
  for(i=0; array[i] != '\0'; i++)
    {
      printf("%c", array[i]);
    }
  printf("\n");
}

/* for printing single phoneme */
void printPhoneme(phoneme *p)
{
  printf("%s - number of frames: %d\n", p->filename, p->frameAmount);
  printf("Id: %d, Cluster: %d, Distortion: %.2f\n", p->phNo, p->myCluster, p->distortion);
  printFrames(p);
}

/* for printing all phonemes */
void printPhonemes(phoneme **phs, int n)
{
  int i;

  for(i=0; i < n; i++)
    printPhoneme(phs[i]);
}

void printPrototypes(phoneme **protos, int k)
{
  int i;
  for (i=0; i < k; i++)
    {
      printf("Prototype of Cluster %d (%d*%d):\n\n", 
	     i, 
	     protos[i]->frameAmount,
	     protos[i]->dimensions);
      printFrames(protos[i]);
    }
}

/* for printing current clustering */
void printClusters(cluster **cs, int m)
{
  int i, j;
  for (i=0; i < m; i++)
    {
      printf("Cluster %d (%.3f) ", cs[i]->cNo, cs[i]->distortion);
      if (cs[i]->exemplar != NULL)
	printf("[%s]", cs[i]->exemplar->filename);
      else  
	printf("[No exemplar]");

      if (cs[i]->phAmount <= 0)
	{
	  printf("\n    No phonemes found.\n\n");
	  continue;
	}

      printf(" #ph: %d\n", cs[i]->phAmount);
      for(j=0; j < cs[i]->phAmount; j++)
	{
	  if ((j % 6) == 0)
	    printf("\n    ");
	  printf("%-10s ",cs[i]->phs[j]->filename);
	}
      
      printf("\n\n");
    }
}

void printClusterSizes(cluster **cs, int m)
{
  int i;
  for(i=0; i < m; i++)
    {
      printf("%d ",cs[i]->phAmount);
    }
  printf("\n");
}

void printUsage()
{
  printf("Time-series clustering\n");
  printf("By Pekka Nykanen\n");
  printf("Department of Computer Science, University of Joensuu\n\n");
  printf("USAGE: tsc [options]\n\n");

  printf("Mandatory options:\n");
  printf("  -i --input [input directory]\n");
  printf("Optional parameters:\n");
  //printf("  -o --output [output directory]\n");
  printf("  -e --extension [file extension]\n");
  printf("  -c --clusters [number of clusters]\n");
  printf("  -m --method [clustering method]\n");
  printf("\t0: K-Medoids\n");
  printf("\t1: Iterative k-medoids\n");
  printf("\t2: Randomised local search\n");
  printf("\t3: Hierarchical clustering\n");
  printf("\t4: K-medoids (with path prototypes)\n");
  printf("\t5: RLS (with path prototypes)\n");
  printf("\t6: HC (with path prototypes)\n");
  printf("\t7: K-medoids (with proposed prototypes)\n");
  printf("\t8: RLS (with proposed prototypes)\n");
  printf("\t9: HC (with proposed prototypes)\n");
  printf("  -r --iterations [RLS-iterations]\n");
  printf("  -h --help [this help]\n");
  exit(-1);
}

/* for printing matrix */
void printMatrix(double **M, int side)
{
  int i, j;
  for(i=0; i < side; i++)
    {
      for(j=0; j < side; j++)
        printf("%-6.1f ", M[i][j]);
      printf("\n");
    }
  printf("\n");
}

/* for printing matrix which is formed by DTW */
void printDTWMatrix(double **M, int rows, int columns)
{
  int i, j;
  for(i=0; i <= rows; i++)
    {
      printf("%-2.2d. ", i);
      for(j=0; j <= columns; j++)
        {
          if (M[i][j] == DBL_MAX)
            printf(" infi. ");
          else if(M[i][j] == 0.0)
            printf("0.000 ");
          else
            printf("%-6.1f ", M[i][j]);
        }
      printf("\n");
    }
}

/* for printing status of clusters */
void printClusterStatus(cluster **cs, int m)
{
  unsigned short i = 0;

  while (i < (m/2))
    {
      printf("Cluster %3.2d (%10s): %-3d    ",
	     cs[i]->cNo, 
	     cs[i]->exemplar->filename, 
	     cs[i]->phAmount);
      if ((m % 2) == 1)
	printf("Cluster %3.2d (%10s): %-3d\n",
	       cs[(m/2) + (i+1)]->cNo, 
	       cs[(m/2) + (i+1)]->exemplar->filename, 
	       cs[(m/2) + (i+1)]->phAmount);
      else
	printf("Cluster %3.2d (%10s): %-3d\n",
	       cs[(m/2) + i]->cNo, 
	       cs[(m/2) + i]->exemplar->filename, 
	       cs[(m/2) + i]->phAmount);
      i++;
    }
  if ((m%2) == 1)
    printf("Cluster %3.2d (%10s): %-3d\n",
	     cs[m/2]->cNo, 
	     cs[m/2]->exemplar->filename, 
	     cs[m/2]->phAmount);
}

void printClusterStatusSimple(cluster **cs, int m)
{
  unsigned short int i;

  for (i=0; i < m; i++)
    {
      if (cs[i]->exemplar != NULL)
	printf("Cluster %3.2d (%10s): %-3d",
	       cs[i]->cNo, 
	       cs[i]->exemplar->filename, 
	       cs[i]->phAmount);
      else
	printf("Merged");
      if ((i % 2) == 0)
	printf("   ");
      else
	printf("\n");
    }
  printf("\n");
}

void printParams(parameters *params)
{
  printf("path:       %s\n",params->path);
  printf("indir:      %s\n",params->indir);
  printf("outdir:     %s\n",params->outdir);
  printf("extension:  %s\n",params->extension);
  printf("clusters:   %d\n",params->clusters);
  printf("method:     %d\n",params->cmethod);
  printf("iterations: %d\n",params->iterations);
  printf("km-iters:   %d\n",params->kmIters);
  printf("one file:   %d\n",params->oneFile);
}

void printResults(double dist, phoneme **phs, phoneme **class, int n, int r, int k)
{
  switch(k)
    {
    case 0:
      if (class != NULL)
	printf("K\t%f\t%f\n", dist, ClassificationError(phs, class, n));
      else
	printf("K\t%f\n", dist);
      break;
    case 1:
      if (class != NULL)
	printf("R-%d\t\t\t%f\t%f\n", r, dist, ClassificationError(phs, class, n));
      else
	printf("R-%d\t\t%f\n", r, dist);
      break;
    case 2:
      if (class != NULL)
	printf("KP\t\t\t\t\t%f\t%f\n", dist, ClassificationError(phs, class, n));
      else
	printf("KP\t\t\t%f\n", dist);
      break;
    case 3:
      if (class != NULL)
	printf("RP-%d\t\t\t\t\t\t\t%f\t%f\n", r, dist, ClassificationError(phs, class, n));
      else
	printf("RP-%d\t\t\t\t%f\n", r, dist);
      break;
    case 4:
      if (class != NULL)
	printf("KE\t\t\t\t\t\t\t\t\t%f\t%f\n", dist, ClassificationError(phs, class, n));
      else
	printf("KE\t\t\t\t\t%f\n", dist);
      break;
    case 5:
      if (class != NULL)
	printf("RE-%d\t\t\t\t\t\t\t\t\t\t\t%f\t%f\n", r, dist, ClassificationError(phs, class, n));
      else
	printf("RE-%d\t\t\t\t\t\t%f\n", r, dist);
      break;
    default:
      break;
    }
}

void printResultsNoLC(double dist, phoneme **phs, phoneme **class, int n)
{
  if (class != NULL)
    printf("%f\t%f\t", dist, ClassificationError(phs, class, n));
  else
    printf("%f\t", dist);
}

void freeParams(parameters *params)
{
  free(params->path);
  free(params->indir);
  free(params->outdir);
  free(params->extension);
  free(params);
}

parameters* initParams()
{
  parameters *param;

  param = (parameters*)malloc(sizeof(parameters));
  param->path = (char*)malloc(MAXPATHLEN*sizeof(char));
  param->indir = (char*)malloc(MAXPATHLEN*sizeof(char));
  param->outdir = (char*)malloc(MAXPATHLEN*sizeof(char));
  param->extension = (char*)malloc(MAXEXTENSIONLEN*sizeof(char));

  if ( (param == NULL)
       || (param->path == NULL) 
       || (param->indir == NULL) 
       || (param->outdir == NULL) 
       || (param->extension == NULL) ) 
    {
      printf("Error while trying to allocate memory for parameters.\n");
      exit(-1);
    }

  getPath(param->path); 
  param->indir[0] = '\0';
  param->outdir[0] = '\0';
  param->extension[0] = '\0';
  param->clusters = 5;
  param->cmethod = 0;
  param->iterations = 1;
  param->kmIters = 0;
  param->oneFile = 0;

  return param;
}

void processParams(parameters *param, int argc, char **argv)
{
  int c, option_index;
  extern char *optarg;
  extern int optind;
  char opt_str[] = "i:o:e:c:m:r:k:fh0";
  char *tmp;

  static struct option long_options[] = {
    { "input",     1, NULL, 'i' },
    { "output",    1, NULL, 'o' },
    { "extension", 1, NULL, 'e' },
    { "clusters",  1, NULL, 'c' },
    { "method",    1, NULL, 'm' },
    { "iterations",1, NULL, 'r' },
    { "km-iters",  1, NULL, 'k' },
    { "file",      0, NULL, 'f' },
    { "help",      0, NULL, 'h' },
    { NULL,        0, NULL, 0 }
  };

  option_index = 0;
  while ((c = getopt_long(argc, argv, opt_str, long_options, &option_index)) != EOF)
    {
      if (c == -1)
        return;

      switch(c) {
      case 'i':
        param->indir = strcpy(param->indir, optarg);
        break;
      case 'o':
        param->outdir = strcpy(param->outdir, optarg);
        break;
      case 'e':
        param->extension = strcpy(param->extension, optarg);
        break;
      case 'c':
        param->clusters = (int) atoi(optarg);
        break;
      case 'm':
        param->cmethod = (int) atoi(optarg);
        break;
      case 'r':
        param->iterations = (int) atoi(optarg);
        break;
      case 'k':
        param->kmIters = (int) atoi(optarg);
        break;
      case 'f':
	param->oneFile = 1;
	break;
      case 'h':
        printUsage();
        exit(0);
        break;
      default:
        printf("Unknown param: %c\n", c);
        printUsage();
        break;
      }
    }

  tmp = (char*)malloc(MAXPATHLEN*sizeof(char));

  if (param->indir[0] == '\0')
    {
      strcpy(param->indir, "ts\0");
      printf("Using default input path: %s/%s\n",param->path,param->indir);
    }
  strcpy(tmp,param->path);
  strcat(tmp,"/");
  strcat(tmp,param->indir);
  strcpy(param->indir,tmp);

  if (param->outdir[0] == '\0')
    {
      strcpy(param->outdir, "out\0");
      //printf("Using default output path: %s/%s\n",param->path,param->outdir);
    }
  strcpy(tmp,param->path);
  strcat(tmp,"/");
  strcat(tmp,param->outdir);
  strcpy(param->outdir,tmp);

  if (param->extension[0] == '\0')
    {
      strcpy(param->extension, ".ph\0");
      //printf("Using default file extension: %s\n",param->extension);
    }

  free(tmp);
}
