/* $Id: structs.c,v 1.12 2008/03/30 23:01:53 pnykanen Exp $ */

#include "structs.h"

/* for allocating memory to single phoneme */
phoneme* newPhoneme(char *name, int dim)
{
  phoneme *p = (phoneme*)malloc(sizeof(phoneme));
  p->filename = (char*)malloc(strlen(name)+1);
  strcpy(p->filename,name);
  p->dimensions = dim;
  p->distortion = DBL_MAX;
  p->frameAmount = 0;
  p->myCluster = -1;
  p->DTWpath = NULL;
  p->frames = NULL;
  return p;
}

/* for allocating memory to all phoneme pointers */
phoneme** newPhonemes(int count)
{
  return (phoneme**)malloc(count*sizeof(phoneme*));
}

void allocateDTWpaths(phoneme **phs, int n, int maxPathLen)
{
  int i,j;
  for(i=0; i < n; i++)
    {
      phs[i]->DTWpath = (char*)calloc(maxPathLen,sizeof(char));
      if (phs[i]->DTWpath == NULL) {
	printf("Error while trying to allocate memory for DTW-paths");
	exit(4);
      }
      for(j=0; j < maxPathLen; j++)
	phs[i]->DTWpath[j] = '\0';
    }
}

/* for allocating memory to single cluster */
cluster* newCluster(int maxph, int cnumber)
{
  int i;
  cluster *c = (cluster*)malloc(sizeof(cluster));
  c->cNo = cnumber;
  c->phAmount = 0;
  c->distortion = 0.0;
  c->exemplar = NULL;
  c->phs = (phoneme**)malloc(maxph*sizeof(phoneme*));
  for (i=0; i < maxph; i++)
    {
      c->phs[i] = NULL;
    }
  return c;
}

/* for allocating memory to all clusters */
cluster** newClusters(int n, int m)
{
  int i;
  cluster **cs = (cluster**)malloc(m*sizeof(cluster*));
  for(i=0; i < m; i++)
    cs[i] = newCluster(n, i);
  return cs;
}

/* for allocating memory to single frame */
double* newFrame(int dim)
{
  return (double*)calloc(dim,sizeof(double));
}

/* for allocating memory to frametable */
double** newFrames(int frAmount, int dim)
{
  int i;
  double **pFrames;

  pFrames = (double**)malloc(frAmount*sizeof(double*));
  if (pFrames == NULL)
    {
      printf("Error while trying to allocate memory for frames!\n");
      exit(-1);
    }

  for (i=0; i < frAmount; i++)
    {
      pFrames[i] = newFrame(dim);
      if (pFrames[i] == NULL)
	{
	  printf("Error while trying to allocate memory for frame!\n");
	  exit(-1);
	}
    }

  return pFrames;
}

char*** newPathMatrix(int n, int l)
{
  int i, j;
  char ***p;

  p = (char***)malloc(n*sizeof(char**));

  for(i=0; i < n; i++) {
    p[i] = (char**)malloc(n*sizeof(char*));
    for(j=0; j < n; j++)
      {
        p[i][j] = (char*)calloc(l,sizeof(char));
      }
  }
  return p;
}

char** newPathVector(int n, int l)
{
  int i;
  char **p;

  p = (char**)malloc(n*sizeof(char*));

  for(i=0; i < n; i++) {
    p[i] = (char*)calloc(l,sizeof(char));
  }
  return p;
}

void freePathMatrix(char ***p, int n, int l)
{
  int i, j;
  for(i=0; i < n; i++) {
    for(j=0; j < n; j++)
      {
	free(p[i][j]);
      }
    free(p[i]);
  }
  free(p);
}

phoneme** newPrototypes(cluster **cs, int m)
{
  char *tmp;
  int i,j;
  phoneme **prototypes;

  tmp = (char*)malloc(MAXPATHLEN*sizeof(char));
  for (j=0; j < MAXPATHLEN; j++)
    tmp[j] = '\0';

  prototypes = newPhonemes(m);
  
  for (i=0, j=0; i < m; i++, j=0)
    {
      sprintf(tmp, "Prototype %d", i);
      prototypes[i] = newPhoneme(tmp,cs[i]->exemplar->dimensions);
      prototypes[i]->phNo = i;
      prototypes[i]->frameAmount = cs[i]->exemplar->frameAmount;
      prototypes[i]->myCluster = i;

      prototypes[i]->frames = newFrames(prototypes[i]->frameAmount, 
					prototypes[i]->dimensions);

      while (tmp[j] != '\0')
	tmp[j++] = '\0';
    }

  free(tmp);

  return prototypes;
}

/* for inserting values into end of frametable of given phoneme */
void insertFrame(phoneme *p, double *values)
{
  int i;

  p->frames = (double**)realloc(p->frames,(p->frameAmount+1)*sizeof(double*));
  p->frames[p->frameAmount] = newFrame(p->dimensions);

  for (i = 0; i < p->dimensions; i++)
    {
      p->frames[p->frameAmount][i] = values[i];
    }
  p->frameAmount += 1;
}

/* for freeing memory used to store clusters */
void freeClusters(cluster **cs, int m)
{
  int i,j;
  for(i=0; i < m; i++)
    {
      for (j=0; j < cs[i]->phAmount; j++)
	cs[i]->phs[j] = NULL;
      free(cs[i]->phs);
      free(cs[i]);
      cs[i] = NULL;
    }
  free(cs);
  cs = NULL;
}

void freePhoneme(phoneme *ph)
{
  if (ph != NULL)
    {
      free(ph->filename);
      ph->filename = NULL;
      free(ph->frames);
      ph->frames = NULL;
      free(ph->DTWpath);
      ph->DTWpath = NULL;
      free(ph);
      ph = NULL;
    }
}

/* for freeing memory used to store phonemes */
void freePhonemes(phoneme **phs, int n)
{
  int i;

  for (i=0; i < n; i++)
    {
      freePhoneme(phs[i]);
    }
  free(phs);
  phs = NULL;
}

void freePrototypes(phoneme **prototypes, int k)
{
  int i,j;

  for (i=0; i < k; i++)
    {
      if (prototypes[i] != NULL)
	{
	  free(prototypes[i]->filename);
	  prototypes[i]->filename = NULL;
	  for (j=0; j < prototypes[i]->frameAmount; j++)
	    {
	      free(prototypes[i]->frames[j]);                   ;
	      prototypes[i]->frames[j] = NULL;
	    }
	  free(prototypes[i]);
	  prototypes[i] = NULL;
	}
    }

  free(prototypes);
}
