/* $Id: tools.c,v 1.10 2008/03/30 23:01:54 pnykanen Exp $ */

#include "tools.h"

char* iToStr(int number)
{
  char *str = NULL;

  str = (char*)malloc(10*sizeof(char));

  sprintf(str, "%d", number);

  return str;
}

double RelDif(double a, double b)
{
  double c = ABS(a);
  double d = ABS(b);

  d = MAX(c, d);

  return d == 0.0 ? 0.0 : ABS(a - b) / d;
}

int countLines(char *filename)
{
  int count = 0;
  int dim = 0, inc = 0, err = 0;
  double *vec;
  FILE *fp = fopen(filename,"r");

  if (fp == NULL)
    {
      printf("Error while trying to read file!\n");
      exit(1);
    }

  vec = (double*)malloc(sizeof(double));

  while ((vec = ReadVector(fp, &dim, &inc, &err)) != NULL)
    count++;

  fclose(fp);

  return count;
}

int checkIfDone(cluster **cs, phoneme **ex, int *sizes, int size)
{
  int i;
  for (i=0; i < size; i++)
    {
      if ( (ex[i] != cs[i]->exemplar) || 
	   (sizes[i] != cs[i]->phAmount) )
        return FALSE;
    }
  return TRUE;
}

int checkIfDoneSizes(cluster **cs, int *sizes, int size)
{
  int i;
  for (i=0; i < size; i++)
    {
      if (sizes[i] != cs[i]->phAmount)
        return FALSE;
    }
  return TRUE;
}

/* for checking index of biggest double in table T */
int biggestOf(double *T, int k)
{
  int i, idx;
  int max;
  if (T == NULL)
    return -1;

  idx = 0;
  max = T[idx];
  for (i=0; i < k; i++)
    {
      if (max < T[i])
	{
	  max = T[i];
	  idx = i;
	}
    }
  return idx;
}

/* for checking index of smallest double in table T */
int smallestOf(double *T, int k)
{
  int i, idx;
  int min;
  if (T == NULL)
    return -1;

  idx = 0;
  min = T[idx];
  for (i=0; i < k; i++)
    {
      if (min > T[i])
	{
	  min = T[i];
	  idx = i;
	}
    }
  return idx;
}

/* for checking if value is in list, used while making random codebook */
int inList(int *list, int size, int value)
{
  int i;

  for (i=0; i < size; i++)
    {
      if (value == list[i])
        return i;
    }
  return -1;
}

int isExemplar(phoneme *ph, cluster **cs, int m)
{
  unsigned int i;
  for(i=0; i < m; i++)
    {
      if (ph == cs[i]->exemplar)
	return TRUE;
    }
  return FALSE;
}

int countMaxFrameSize(phoneme **phs, int n)
{
  int i, curr_max = 0;
  for(i = 0; i < n; i++)
    {
      if (phs[i]->frameAmount > curr_max)
	curr_max = phs[i]->frameAmount;
    }
  return curr_max;
}

void addFrameValues(double *proframe, double *frame, int dim)
{
  int i;

  for (i = 0; i < dim; i++)
    {
      proframe[i] += frame[i];
    }
}

void clearPrototype(double **frames, int n, int m)
{
  int i, j;

  for (i=0; i < n; i++)
    for (j=0; j < m; j++)
      frames[i][j] = 0.0;
}

double** reallocateFrames(phoneme *target, int newRows, int newColumns)
{
  int i;
  double **newFrames = NULL;

  if (target != NULL)
    {
      if (target->frameAmount > 0)
	for (i=0; i < target->frameAmount; i++)
	  free(target->frames[i]);
      free(target->frames);

      newFrames = (double**)malloc(newRows*sizeof(double*));
      if (newFrames == NULL)
	{
	  printf("Error while trying to reallocate new rows\n");
	  exit(1);
	}

      for(i=0; i < newRows; i++)
	{
	  newFrames[i] = (double*)calloc(newColumns,sizeof(double));      
	  if (newFrames[i] == NULL)
	    {
	      printf("Error while trying to reallocate new columns\n");
	      exit(2);
	    }
	}
      target->frameAmount = newRows;
    }

  return newFrames;
}

void formMeanPhoneme(phoneme *proto, cluster *cs)
{
  int i, j, k, l;
  int *divs;

  clearPrototype(proto->frames, proto->frameAmount, proto->dimensions);

  if (proto->frameAmount != cs->exemplar->frameAmount)
    {
      proto->frames = reallocateFrames(proto,cs->exemplar->frameAmount,cs->exemplar->dimensions);
      if (proto->frames == NULL)
	{
	  printf("Error while reallocating memory for prototype's frames!\n");
	  exit(1);
	}
    }

  divs = (int*)calloc(proto->frameAmount,sizeof(int));
  if (divs == NULL)
    {
      printf("Error while allocating memory for prototype's dividers!\n");
      exit(2);
    }

  for (i = 0; i < cs->phAmount; i++) // for each phoneme in a cluster
    {
      j = -1; k = -1; l = 0;
      while ( (j < cs->phs[i]->frameAmount - 1) || (k < cs->exemplar->frameAmount - 1) )
	{
	  switch (cs->phs[i]->DTWpath[l]) // move to next cell in DTW-path
	    {
	    case '1': 
	      j++;
	      break;
	    case '2': 
	      k++;
	      break;
	    case '3': 
	      j++; k++;
	      break;
	    }

	  divs[k] += 1;

	  addFrameValues(proto->frames[k], cs->phs[i]->frames[j], proto->dimensions);
	  	  
	  l++;
	}
    }
  
  // divide each cell by phoneme amount to make mean representation
  // of phonemes in this cluster
  for (i = 0; i < proto->frameAmount; i++)
    {
      for (j = 0; j < proto->dimensions; j++)
	proto->frames[i][j] /= (double)divs[i];      
    }

  free(divs);
}

void invertDTWpath(char *array)
{
  char tmp;
  int i, j;

  // forward to 0
  i=0;
  while (array[i] != '\0') i++;
  i--;

  for (j=0; j < i; i--, j++)
    {
      tmp = array[j];
      array[j] = array[i];
      array[i] = tmp;
    }
}

void copyPaths(char *a, char *b)
{
  int i=0;
  while (b[i] != '\0') { 
    a[i] = b[i]; 
    i++;
  }
  
  while (a[i] != '\0')
    a[i] = '\0';
}

void copyAndFlipPaths(char *a, char *b)
{
  int i=0;
  while (b[i] != '\0') {
    if (b[i] == '1')
      a[i] = '2';
    if (b[i] == '2')
      a[i] = '1';
    else
      a[i] = b[i]; 
    i++;
  }
  
  while (a[i] != '\0')
    a[i] = '\0';
}

void copyCluster(cluster *target, cluster *source)
{
  int i;

  target->cNo = source->cNo;
  target->phAmount = source->phAmount;
  target->distortion = source->distortion;
  target->exemplar = source->exemplar;
  for(i=0; i < source->phAmount; i++)
    {
      target->phs[i] = source->phs[i];
    }
}

/* for copying clusters, used to store the best found solution for given set */
void copyClusters(cluster **target, cluster **source, int m)
{
  int i;
  
  for(i=0; i < m; i++)
    {
      copyCluster(target[i],source[i]);
    }
}

void copyPhAndCluster(cluster *target, cluster *source)
{
  int i;

  target->cNo = source->cNo;
  target->phAmount = source->phAmount;
  target->distortion = source->distortion;
  target->exemplar = source->exemplar;
  for(i=0; i < source->phAmount; i++)
    {
      target->phs[i] = source->phs[i];
      target->phs[i]->myCluster = target->cNo;
    }
}

/* for copying phs & clusters, used to store the best found solution for given set */
void copyPhsAndClusters(cluster **target, cluster **source, int m)
{
  int i;
  
  for(i=0; i < m; i++)
    {
      copyPhAndCluster(target[i],source[i]);
    }
}

void copyPrototype(phoneme *target, phoneme *source)
{
  int i,j;
  
  if (target != NULL)
    {
      if (target->frameAmount != source->frameAmount)
	target->frames = reallocateFrames(target, source->frameAmount, source->dimensions);
      
      target->phNo = source->phNo;
      target->dimensions = source->dimensions;
      target->myCluster = source->myCluster;
      target->distortion = source->distortion;
      
      i = 0;
      while (source->filename[i] != '\0')
	{
	  target->filename[i] = source->filename[i];
	  i++;
	}
      target->filename[i] = '\0';
      
      for(i=0; i < source->frameAmount; i++)
	for(j=0; j < source->dimensions; j++)
	  target->frames[i][j] = source->frames[i][j];
    }
}

void copyPrototypes(phoneme **targets, phoneme **sources, int n)
{
  int i;

  for (i=0; i < n; i++)
    {
      copyPrototype(targets[i],sources[i]);
    }
}

void resetCluster(cluster *cs)
{
  int i;
  for (i=0; i < cs->phAmount; i++) 
    {
      cs->phs[i] = NULL;
    }
  cs->phAmount = 0;
  cs->distortion = 0.0;
  cs->exemplar = NULL;
}

/* for reseting clustering, used while experimenting with different cluster sizes */
void resetClusters(cluster **cs, int m)
{
  int i;
  
  for(i=0; i < m; i++)                            // for each cluster
    {
      resetCluster(cs[i]);
    }
}

void resetPhsAndCluster(cluster *cs)
{
  int i;

  for (i=0; i < cs->phAmount; i++) 
    {
      cs->phs[i]->distortion = DBL_MAX;
      cs->phs[i]->myCluster = -1;
      cs->phs[i] = NULL;
    }
  cs->phAmount = 0;
  cs->distortion = 0.0;
  cs->exemplar = NULL;
}

/* for reseting clustering, used while experimenting with different cluster sizes */
void resetPhsAndClusters(cluster **cs, int m)
{
  int i;
  
  for(i=0; i < m; i++)                            
    { 
      resetPhsAndCluster(cs[i]);
    }
}

void resetPrototype(phoneme *proto)
{
  int i,j;

  if (proto != NULL)
    {
      for(i=0; i < proto->frameAmount; i++)
	for(j=0; j < proto->dimensions; j++)
	  proto->frames[i][j] = 0.0;
      
      proto->myCluster = -1;
      proto->distortion = 0.0;
    }
}

void resetPrototypes(phoneme **protos, int n)
{
  int i;

  for (i=0; i < n; i++)
    {
      resetPrototype(protos[i]);
      protos[i]->phNo = i;
      protos[i]->myCluster = i;
    }
}

/* for getting random codebook which is used in start of clustering algorithms */
void getRandomCodebook(phoneme **phs, cluster **cs, int n, int m)
{
  int i; 
  int browser = 0;
  int chosenPhoneme = -1;
  int *listOfChosen = (int*)malloc(m*sizeof(int));
  for (i=0; i < m; i++)
    {
      do
        {
          chosenPhoneme = (int)(((double)n)*rand()/(RAND_MAX)); 
	  //(((unsigned)rand()) % n); 
        }
      while (inList(listOfChosen,browser,chosenPhoneme) >= 0);

      listOfChosen[browser++] = chosenPhoneme;
      cs[i]->exemplar = phs[chosenPhoneme];
    }

  free(listOfChosen);
}

void createClassification(phoneme **phs, int n)
{
  int i, tmp;

  switch(n)
    {
    case 600: // synthetic
      for (i=0; i < n; i++)
	{
	  tmp = i / 100;
	  switch(tmp)
	    {
	    case 0:
	      phs[i]->myCluster = 0;
	      break;
	    case 1:
	      phs[i]->myCluster = 1;
	      break;
	    case 2:
	      phs[i]->myCluster = 2;
	      break;
	    case 3:
	      phs[i]->myCluster = 3;
	      break;
	    case 4:
	      phs[i]->myCluster = 4;
	      break;
	    case 5:
	      phs[i]->myCluster = 5;
	      break;
	    }
	}
      break;
    case 640: // japanese
      for (i=0; i < n; i++)
	{
	  phs[i]->myCluster = atoi(phs[i]->filename) - 1;
	}
      break;
    case 925: // character
      for (i=0; i < n; i++)
	{
	  switch(phs[i]->filename[4])
	    {
	    case 'a': phs[i]->myCluster = 0;
	      break;
	    case 'A': phs[i]->myCluster = 1;
	      break;
	    case 'b': phs[i]->myCluster = 2;
	      break;
	    case 'B': phs[i]->myCluster = 3;
	      break;
	    case 'c': phs[i]->myCluster = 4;
	      break;
	    case 'C': phs[i]->myCluster = 5;
	      break;
	    case 'd': phs[i]->myCluster = 6;
	      break;
	    case 'D': phs[i]->myCluster = 7;
	      break;
	    case 'e': phs[i]->myCluster = 8;
	      break;
	    case 'E': phs[i]->myCluster = 9;
	      break;
	    case 'f': phs[i]->myCluster = 10;
	      break;
	    case 'F': phs[i]->myCluster = 11;
	      break;
	    case 'g': phs[i]->myCluster = 12;
	      break;
	    case 'G': phs[i]->myCluster = 13;
	      break;
	    case 'h': phs[i]->myCluster = 14;
	      break;
	    case 'H': phs[i]->myCluster = 15;
	      break;
	    case 'i': phs[i]->myCluster = 16;
	      break;
	    case 'I': phs[i]->myCluster = 17;
	      break;
	    case 'j': phs[i]->myCluster = 18;
	      break;
	    case 'J': phs[i]->myCluster = 19;
	      break;
	    case 'k': phs[i]->myCluster = 20;
	      break;
	    case 'K': phs[i]->myCluster = 21;
	      break;
	    case 'l': phs[i]->myCluster = 22;
	      break;
	    case 'L': phs[i]->myCluster = 23;
	      break;
	    case 'm': phs[i]->myCluster = 24;
	      break;
	    case 'M': phs[i]->myCluster = 25;
	      break;
	    case 'n': phs[i]->myCluster = 26;
	      break;
	    case 'N': phs[i]->myCluster = 27;
	      break;
	    case 'o': phs[i]->myCluster = 28;
	      break;
	    case 'O': phs[i]->myCluster = 29;
	      break;
	    case 'p': phs[i]->myCluster = 30;
	      break;
	    case 'P': phs[i]->myCluster = 31;
	      break;
	    case 'q': phs[i]->myCluster = 32;
	      break;
	    case 'Q': phs[i]->myCluster = 33;
	      break;
	    case 'r': phs[i]->myCluster = 34;
	      break;
	    case 'R': phs[i]->myCluster = 35;
	      break;
	    case 's': phs[i]->myCluster = 36;
	      break;
	    case 'S': phs[i]->myCluster = 37;
	      break;
	    case 't': phs[i]->myCluster = 38;
	      break;
	    case 'T': phs[i]->myCluster = 39;
	      break;
	    case 'u': phs[i]->myCluster = 40;
	      break;
	    case 'U': phs[i]->myCluster = 41;
	      break;
	    case 'v': phs[i]->myCluster = 42;
	      break;
	    case 'V': phs[i]->myCluster = 43;
	      break;
	    case 'w': phs[i]->myCluster = 44;
	      break;
	    case 'W': phs[i]->myCluster = 45;
	      break;
	    case 'y': phs[i]->myCluster = 46;
	      break;
	    case 'Y': phs[i]->myCluster = 47;
	      break;
	    case 'z': phs[i]->myCluster = 48;
	      break;
	    case 'Z': phs[i]->myCluster = 49;
	      break;
	    case '0': phs[i]->myCluster = 50;
	      break;
	    case '1': phs[i]->myCluster = 51;
	      break;
	    case '2': phs[i]->myCluster = 52;
	      break;
	    case '3': phs[i]->myCluster = 53;
	      break;
	    case '4': phs[i]->myCluster = 54;
	      break;
	    case '5': phs[i]->myCluster = 55;
	      break;
	    case '6': phs[i]->myCluster = 56;
	      break;
	    case '8': phs[i]->myCluster = 57;
	      break;
	    case '9': phs[i]->myCluster = 58;
	      break;
	    default:
	      break;
	    }
	}
      break;
    case 2565: // auslan
      for (i=0, tmp = 0; i < n; i += 3)
	{
	  phs[i]->myCluster = tmp;
	  phs[i+1]->myCluster = tmp;
	  phs[i+2]->myCluster = tmp;
	  if (tmp < 94)
	    tmp++;
	  else
	    tmp = 0;
	}
      break;
    default: 
      printf("Unknown amount of time-series for classification\n");
      exit(1);
    }

}
