/* $Id: tsc.c,v 1.13 2008/03/30 23:01:54 pnykanen Exp $ */

#include "tsc.h"

/* for assigning each phoneme to it's 'currently nearest' cluster */
void assignClusters(phoneme **phs, cluster **cs, int n, int m)
{
  int i,j;
  int dest_c = 0;
  double curr_DTW;
  double curr_min = DBL_MAX;

  for (i=0; i < n; i++, curr_min = DBL_MAX)          // for each phoneme
    {
      for (j=0; j < m; j++)                             // for each cluster
        {
	  curr_DTW = DTW(phs[i], cs[j]->exemplar);
          if (curr_min > curr_DTW)
            {
              dest_c = j;
              curr_min = curr_DTW;
            }
        }

      phs[i]->distortion = curr_min;     // update distortion

      if (phs[i]->myCluster == dest_c)   // if this phoneme is staying in cluster
	continue;

      if (phs[i]->myCluster >= 0)    // remove from old cluster if not 1st time
	{ 
	  // rewind
	  for (j=0; cs[phs[i]->myCluster]->phs[j] != phs[i]; j++) { }
	  
	  /* swap last to current place, empty last and decrease phAmount */
	  cs[phs[i]->myCluster]->phs[j] = cs[phs[i]->myCluster]->phs[cs[phs[i]->myCluster]->phAmount-1];
	  cs[phs[i]->myCluster]->phs[cs[phs[i]->myCluster]->phAmount-1] = NULL;	  
	  cs[phs[i]->myCluster]->phAmount -= 1;
	}

      phs[i]->myCluster = dest_c;                        // assign to new cluster
      cs[dest_c]->phs[cs[dest_c]->phAmount] = phs[i];
      cs[dest_c]->phAmount++;
    }
}

void assignMedoid(cluster *cs)
{
  int j, k, min_idx;
  double **dist_matrix = NULL;
  double *sum_table = NULL;
  
  dist_matrix = getDistanceMatrix(cs->phs, cs->phAmount);
  sum_table = (double*)calloc(cs->phAmount,sizeof(double)); 
  
  for(j=0; j < cs->phAmount; j++) // each phoneme in cluster
    {
      for(k=0; k < cs->phAmount; k++) // each other phoneme in cluster
	{
	  sum_table[j] += dist_matrix[j][k];
	}
    }
  
  min_idx = smallestOf(sum_table, cs->phAmount);
  cs->exemplar = cs->phs[min_idx];

  calculateIntDist(cs, sum_table[min_idx]);

  for(j=0; j < cs->phAmount; j++) // each phoneme in cluster
    {
      //cs->phs[j]->distortion = dist_matrix[j][min_idx];
      free(dist_matrix[j]);
    }
  free(dist_matrix);
  free(sum_table);
}

/* for assigning new medoids for clusters, based on phonemes belonging to each cluster */
void assignMedoids(cluster **cs, int m)
{
  int i;
  
  for(i=0; i < m ; i++)          // for each cluster
    {
      assignMedoid(cs[i]);
    }
}

/* for assigning new medoids for clusters, 
   based on DTW-paths */
void assignMedoid_Prototype(cluster *cs, int plen)
{
  int i, min_idx;
  double *dist_vector = NULL;
  char **paths = NULL;

  paths = newPathVector(cs->phAmount,plen);
    
  dist_vector = getDistanceVectorWithPath(cs->phs, cs->phAmount, &min_idx, paths, plen);

  cs->exemplar = cs->phs[min_idx];

  for(i=0; i < cs->phAmount; i++)
    {
      copyPaths(cs->phs[i]->DTWpath,paths[i]);
      invertDTWpath(cs->phs[i]->DTWpath); // invert found DTW-path      
      free(paths[i]);
    }
  free(dist_vector);
  free(paths);
}

void assignMedoids_Prototype(cluster **cs, int m, int maxpath)
{
  int i;

  for(i=0; i < m ; i++)          // for each cluster
    {
      assignMedoid_Prototype(cs[i], maxpath);
    }
}

void assignDTWpaths(phoneme **phs, cluster **cs, int n)
{
  int i;
  double tmp;

  for (i=0, tmp = DBL_MAX; i < n; i++, tmp = DBL_MAX)
    {
      DTWpath(phs[i], cs[phs[i]->myCluster]->exemplar, &tmp);
      phs[i]->distortion = tmp;
      invertDTWpath(phs[i]->DTWpath);
    }
}

void assignPathsForCluster(phoneme **phs, phoneme *proto, int n)
{
  int i;
  double tmp;

  for (i=0, tmp = DBL_MAX; i < n; i++, tmp = DBL_MAX)
    {
      DTWpath(phs[i], proto, &tmp);
      phs[i]->distortion = tmp;
      invertDTWpath(phs[i]->DTWpath);
    }
}

void updatePrototype(phoneme *proto, cluster *cs)
{
  if (cs->phAmount > 0)
    formMeanPhoneme(proto,cs);
  else
    freePhoneme(proto);
    
  cs->exemplar = proto;  
}

void updatePrototypes(phoneme **protos, cluster **cs, int m)
{
  int i;

  for(i=0; i < m; i++)          // for each cluster
    {
      updatePrototype(protos[i], cs[i]);
    }
}

double kMedoids(phoneme **phs, cluster **cs, int n, int k, int *iters)
{
  int i;
  double currDist, exDist;

  if (*iters != 0)
    {
      for(i=0; i < *iters; i++) {
	assignClusters(phs, cs, n, k);
	assignMedoids(cs, k);
      }
      currDist = calculateDistortion(n,cs,k);
    }
  else 
    {
      currDist = DBL_MAX;
      i = 0;
      do
	{
	  i++;

	  assignClusters(phs, cs, n, k);
	  assignMedoids(cs, k);
	  
	  exDist = currDist;
	  currDist = calculateExternalDistortion(n,cs,k);
	}
      while (RelDif(currDist,exDist) > TOLERANCE);
      *iters = i;
    }
  
  return currDist;
}

/* for creating clustering, using iterative k-medoids */
double iter_kMedoids(phoneme **phs, cluster **cs, int n, int m, int iterations, int *kmIters)
{
  int i;
  double curr_best_dist = DBL_MAX;
  double curr_dist = 0;
  cluster **curr_best = newClusters(n, m);

  for(i = 0; i < iterations; i++)
    {
      if (i > 0) 
        resetPhsAndClusters(cs, m);

      getRandomCodebook(phs, cs, n, m);
      curr_dist = kMedoids(phs, cs, n, m, kmIters);
      
      if (curr_best_dist > curr_dist)
        {
          if (i > 0)
	    resetClusters(curr_best, m);
          copyClusters(curr_best, cs, m);
          curr_best_dist = curr_dist;
        }
    }
  
  resetPhsAndClusters(cs, m);
  copyPhsAndClusters(cs, curr_best, m);

  return curr_best_dist;
}

/* for creating clustering, using iterative k-medoids with prototypes */
double iter_kMedoids_Prototype(phoneme **phs, cluster **cs, phoneme **protos, int n, int m, int iterations, int plen, int *kmIters)
{
  int i;
  double curr_best_dist = DBL_MAX;
  double curr_dist = 0;
  cluster **curr_best = newClusters(n, m);
  phoneme **bestProtos = newPrototypes(cs,m);

  for(i = 0; i < iterations; i++)
    {
      if (i > 0) 
	{
	  resetPhsAndClusters(cs, m);
	  resetPrototypes(protos, m);
	}
      
      getRandomCodebook(phs, cs, n, m);
      curr_dist = kMedoids_Prototype(phs, cs, protos, n, m, plen, kmIters);
      
      if (curr_best_dist > curr_dist)
        {
          if (i > 0)
	    {
	      resetClusters(curr_best, m);
	      resetPrototypes(protos,m);
	    }
	  copyClusters(curr_best, cs, m);
	  copyPrototypes(bestProtos, protos, m);
          curr_best_dist = curr_dist;
        }
    }
  
  resetPhsAndClusters(cs, m);
  copyPhsAndClusters(cs, curr_best, m);
  resetPrototypes(protos, m);
  copyPrototypes(protos, bestProtos, m);

  return curr_best_dist;
}

/* for creating clustering, using iterative k-medoids with proposed prototypes */
double iter_kMedoids_Proposed(phoneme **phs, cluster **cs, phoneme **protos, int n, int m, int iterations, int plen, int *kmIters)
{
  int i;
  double curr_best_dist = DBL_MAX;
  double curr_dist = 0;
  cluster **curr_best = newClusters(n, m);
  phoneme **bestProtos = newPrototypes(cs,m);

  for(i = 0; i < iterations; i++)
    {
      if (i > 0) 
	{
	  resetPhsAndClusters(cs, m);
	  resetPrototypes(protos, m);
	}
      
      getRandomCodebook(phs, cs, n, m);
      curr_dist = kMedoids_Proposed(phs, cs, protos, n, m, plen, kmIters);
      
      if (curr_best_dist > curr_dist)
        {
          if (i > 0)
	    {
	      resetClusters(curr_best, m);
	      resetPrototypes(protos,m);
	    }
	  copyClusters(curr_best, cs, m);
	  copyPrototypes(bestProtos, protos, m);
          curr_best_dist = curr_dist;
        }
    }
  
  resetPhsAndClusters(cs, m);
  copyPhsAndClusters(cs, curr_best, m);
  resetPrototypes(protos, m);
  copyPrototypes(protos, bestProtos, m);

  return curr_best_dist;
}

/* for creating clustering, using k-medoids and protos, saving iterationAmount*/
double kMedoids_Prototype(phoneme **phs, cluster **cs, phoneme **prototypes, int n, int k, int plen, int *iters)
{
  int i;
  double currDist, exDist;

  if (*iters != 0)
    {
      for (i=0; i < *iters; i++) 
	{
	  assignClusters(phs, cs, n, k);
	  assignMedoids_Prototype(cs, k, plen);
	  updatePrototypes(prototypes, cs, k);
	}      	
      currDist = calculateDistortion(n,cs,k);
    }
  else
    {
      i = 0;
      currDist = DBL_MAX;

      do
	{
	  i++;
	  /* count stuff */
	  assignClusters(phs, cs, n, k);
	  assignMedoids_Prototype(cs, k, plen);
	  updatePrototypes(prototypes, cs, k);
	
	  exDist = currDist;
	  currDist = calculateExternalDistortion(n,cs,k);
	}
      while (RelDif(currDist,exDist) > TOLERANCE);

      *iters = i;
    }

  return currDist;
}

/* for creating clustering, using k-medoids and protos */
double kMedoids_Proposed(phoneme **phs, cluster **cs, phoneme **prototypes, int n, int k, int plen, int *iters)
{
  int i,j,q;
  double currDist, exDist, exProtoDist;

  if (*iters != 0)
    {
      for(i=0; i < *iters; i++)
	{
	  assignClusters(phs, cs, n, k);
	  assignMedoids_Prototype(cs, k, plen);
	  updatePrototypes(prototypes, cs, k);
	  
	  for (j=0,q=0; j < k; j++, q=0)
	    {
	      exProtoDist = DBL_MAX;
	      calculateInternalDistortion(cs[j]);
	      
	      do
		{
		  q++;
		  assignPathsForCluster(cs[j]->phs, cs[j]->exemplar, cs[j]->phAmount);
		  updatePrototype(prototypes[j], cs[j]);
		  exProtoDist = cs[j]->distortion;
		  calculateInternalDistortion(cs[j]);
		}
	      while ((RelDif(cs[j]->distortion, exProtoDist) > TOLERANCE) && (q < 2*(*iters)));	      
	    }
	}
      currDist = calculateDistortion(n,cs,k);
    }
  else 
    {
      i = 0;
      currDist = DBL_MAX;
      
      do
	{
	  i++;
	  assignClusters(phs, cs, n, k);
	  assignMedoids_Prototype(cs, k, plen);
	  updatePrototypes(prototypes, cs, k);
	  
	  for (j=0; j < k; j++)
	    {
	      exProtoDist = DBL_MAX;
	      calculateInternalDistortion(cs[j]);
	      
	      do
		{
		  assignPathsForCluster(cs[j]->phs, cs[j]->exemplar, cs[j]->phAmount);
		  updatePrototype(prototypes[j], cs[j]);
		  
		  exProtoDist = cs[j]->distortion;
		  calculateInternalDistortion(cs[j]);
		}
	      while (RelDif(cs[j]->distortion, exProtoDist) > TOLERANCE);
	    }
	  
	  exDist = currDist;
	  currDist = calculateDistortion(n,cs,k);
	}
      while (RelDif(currDist,exDist) > TOLERANCE);
      *iters = i;
    }

  return currDist;
}

/* Swapping of cluster exemplar for RLS */
void swapCluster(phoneme **phs, cluster **cs, int n, int m, int *from, int *to)
{
  int chosenPhoneme;

  *from = (int) (((double)m)*rand()/(RAND_MAX));

  do
    {
      *to = (int) (((double)m)*rand()/(RAND_MAX));
    } 
  while ( ((cs[*to]->phAmount <= 1) || (*from == *to))
	  && (m < n) && (m > 1));

  do
    {
      chosenPhoneme = (int) (((double)cs[*to]->phAmount)*rand()/(RAND_MAX));
    } 
  while ( (cs[*to]->exemplar == cs[*to]->phs[chosenPhoneme])
	  && (m < n) );

  cs[*from]->exemplar = cs[*to]->phs[chosenPhoneme];
}

/* Swapping of cluster's prototype for RLS_Prototype */
void swapPrototype(phoneme **phs, cluster **cs, phoneme **protos, int n, int m, int *from, int *to)
{
  double dbltmp;
  int chosenPhoneme;

  *from = (int) (((double)m)*rand()/(RAND_MAX));

  do
    {
      *to = (int) (((double)m)*rand()/(RAND_MAX));
    } 
  while ( ((cs[*to]->phAmount <= 1) || (*from == *to))
	  && (m < n) && (m > 1) );

  do
    {
      chosenPhoneme = (int) (((double)cs[*to]->phAmount)*rand()/(RAND_MAX));
    } 
  while ( (cs[*to]->exemplar == cs[*to]->phs[chosenPhoneme]) 
	  && (m < n) );

  resetPhsAndCluster(cs[*from]);
  resetPrototype(protos[*from]);
  
  dbltmp = DBL_MAX;
  cs[*from]->phs[0] = cs[*to]->phs[chosenPhoneme];
  cs[*from]->phs[0]->myCluster = *from;
  cs[*from]->phAmount = 1;
  cs[*from]->exemplar = cs[*from]->phs[0];
  DTWpath(cs[*from]->phs[0], cs[*from]->exemplar, &dbltmp);
  cs[*from]->phs[0]->distortion = dbltmp;

  if (m != 1)
    {
      cs[*to]->phs[chosenPhoneme] = cs[*to]->phs[cs[*to]->phAmount-1];
      cs[*to]->phs[cs[*to]->phAmount-1] = NULL;
      cs[*to]->phAmount -= 1;
    }

  updatePrototype(protos[*from],cs[*from]);
}

/* RLS */
double RLS(phoneme **phs, cluster **cs, int n, int m, int r, int *iters)
{
  int done = 0, from = -1, to = -1;
  double curr_best_dist = DBL_MAX;
  double curr_dist = DBL_MAX;
  cluster **curr_best = newClusters(n, m);

  curr_best_dist = kMedoids(phs,cs,n,m, iters);
  copyClusters(curr_best, cs, m);
  //  printf("%f\t",curr_best_dist);

  while (done <= r)
    {
      swapCluster(phs,cs,n,m,&from,&to);
      curr_dist = kMedoids(phs, cs, n, m, iters);
      if (curr_best_dist > curr_dist)
        {
          resetPhsAndClusters(curr_best, m);
	  copyPhsAndClusters(curr_best, cs, m);
	  curr_best_dist = curr_dist;
	}
      else
	{
	  resetPhsAndClusters(cs, m);
	  copyPhsAndClusters(cs, curr_best, m);
	}
      //      printf("%f\t",curr_best_dist);
      done++;
    }

  resetPhsAndClusters(cs, m);
  copyPhsAndClusters(cs, curr_best, m);

  freeClusters(curr_best,m);

  return curr_best_dist;
}

double RLS_Prototype(phoneme **phs, cluster **cs, phoneme **protos, int n, int m, int r, int plen, int *iters)
{
  int done = 0, from = -1, to = -1;
  double curr_best_dist = DBL_MAX;
  double curr_dist = DBL_MAX;
  phoneme **curr_best_protos = newPrototypes(cs,m);
  cluster **curr_best = newClusters(n, m);

  curr_best_dist = kMedoids_Prototype(phs,cs,protos,n,m,plen,iters);
  copyClusters(curr_best, cs, m);
  copyPrototypes(curr_best_protos, protos, m);

  while (done < r)
    {
      swapPrototype(phs,cs,protos,n,m,&from,&to);
      curr_dist = kMedoids_Prototype(phs, cs, protos, n, m, plen, iters);
      if (curr_best_dist > curr_dist)
        {
          resetPhsAndClusters(curr_best, m);
	  copyPhsAndClusters(curr_best, cs, m);
	  resetPrototypes(curr_best_protos, m);
	  copyPrototypes(curr_best_protos, protos, m);
	  curr_best_dist = curr_dist;
	}
      else
	{
	  resetPhsAndClusters(cs, m);
	  copyPhsAndClusters(cs, curr_best, m);
	  resetPrototypes(protos, m);
	  copyPrototypes(protos, curr_best_protos, m);
	}
      done++;
    }

  resetPhsAndClusters(cs, m);
  copyPhsAndClusters(cs, curr_best, m);
  resetPrototypes(protos, m);
  copyPrototypes(protos, curr_best_protos, m);

  freeClusters(curr_best,m);
  freePrototypes(curr_best_protos,m);

  return curr_best_dist;
}

double RLS_Proposed(phoneme **phs, cluster **cs, phoneme **protos, int n, int m, int r, int plen, int *iters)
{
  int done = 0, from = -1, to = -1;
  double curr_best_dist = DBL_MAX;
  double curr_dist = DBL_MAX;
  phoneme **curr_best_protos = newPrototypes(cs,m);
  cluster **curr_best = newClusters(n, m);

  curr_best_dist = kMedoids_Proposed(phs,cs,protos,n,m,plen, iters);
  copyClusters(curr_best, cs, m);
  copyPrototypes(curr_best_protos, protos, m);

  while (done < r)
    {
      swapPrototype(phs,cs,protos,n,m,&from,&to);
      curr_dist = kMedoids_Proposed(phs, cs, protos, n, m, plen, iters);
      if (curr_best_dist > curr_dist)
        {
          resetPhsAndClusters(curr_best, m);
	  copyPhsAndClusters(curr_best, cs, m);
	  resetPrototypes(curr_best_protos, m);
	  copyPrototypes(curr_best_protos, protos, m);
	  curr_best_dist = curr_dist;
	}
      else
	{
	  resetPhsAndClusters(cs, m);
	  copyPhsAndClusters(cs, curr_best, m);
	  resetPrototypes(protos, m);
	  copyPrototypes(protos, curr_best_protos, m);
	}
      done++;
    }

  resetPhsAndClusters(cs, m);
  copyPhsAndClusters(cs, curr_best, m);
  resetPrototypes(protos, m);
  copyPrototypes(protos, curr_best_protos, m);

  freeClusters(curr_best,m);
  freePrototypes(curr_best_protos,m);

  return curr_best_dist;
}

/* RLS with k-medoids */
double RLS_Class(phoneme **phs, cluster **cs, phoneme **class, int n, int m, int r, int *iters)
{
  int done = 0, from = -1, to = -1;
  double curr_best_dist = DBL_MAX;
  double curr_dist = DBL_MAX;
  cluster **curr_best = newClusters(n, m);

  curr_best_dist = kMedoids(phs,cs,n,m,iters);
  copyClusters(curr_best, cs, m);
  //printResultsNoLC(curr_best_dist, phs, class, n);

  while (done < r)
    {
      swapCluster(phs,cs,n,m,&from,&to);
      curr_dist = kMedoids(phs,cs,n,m,iters);
      if (curr_best_dist > curr_dist)
        {
          resetPhsAndClusters(curr_best, m);
	  copyPhsAndClusters(curr_best, cs, m);
	  curr_best_dist = curr_dist;
	}
      else
	{
	  resetPhsAndClusters(cs, m);
	  copyPhsAndClusters(cs, curr_best, m);
	}
      done++;
    }

  resetPhsAndClusters(cs, m);
  copyPhsAndClusters(cs, curr_best, m);

  freeClusters(curr_best,m);

  return curr_best_dist;
}

double RLS_Prototype_Class(phoneme **phs, cluster **cs, phoneme **protos, phoneme **class, int n, int m, int r, int plen, int *iters)
{
  int done = 0, from = -1, to = -1;
  double curr_best_dist = DBL_MAX;
  double curr_dist = DBL_MAX;
  phoneme **curr_best_protos = newPrototypes(cs,m);
  cluster **curr_best = newClusters(n, m);

  curr_best_dist = kMedoids_Prototype(phs,cs,protos,n,m,plen,iters);
  copyClusters(curr_best, cs, m);
  copyPrototypes(curr_best_protos, protos, m);
  //printResultsNoLC(curr_best_dist, phs, class, n);

  while (done < r)
    {
      swapPrototype(phs,cs,protos,n,m,&from,&to);
      curr_dist = kMedoids_Prototype(phs,cs,protos,n,m,plen,iters);
      if (curr_best_dist > curr_dist)
        {
          resetPhsAndClusters(curr_best, m);
	  copyPhsAndClusters(curr_best, cs, m);
	  resetPrototypes(curr_best_protos, m);
	  copyPrototypes(curr_best_protos, protos, m);
	  curr_best_dist = curr_dist;
	}
      else
	{
	  resetPhsAndClusters(cs, m);
	  copyPhsAndClusters(cs, curr_best, m);
	  resetPrototypes(protos, m);
	  copyPrototypes(protos, curr_best_protos, m);
	}
      done++;
    }

  resetPhsAndClusters(cs, m);
  copyPhsAndClusters(cs, curr_best, m);
  resetPrototypes(protos, m);
  copyPrototypes(protos, curr_best_protos, m);

  freeClusters(curr_best,m);
  freePrototypes(curr_best_protos,m);

  return curr_best_dist;
}

double RLS_Proposed_Class(phoneme **phs, cluster **cs, phoneme **protos, phoneme **class, int n, int m, int r, int plen, int *iters)
{
  int done = 0, from = -1, to = -1;
  double curr_best_dist = DBL_MAX;
  double curr_dist = DBL_MAX;
  phoneme **curr_best_protos = newPrototypes(cs,m);
  cluster **curr_best = newClusters(n, m);

  curr_best_dist = kMedoids_Proposed(phs,cs,protos,n,m,plen,iters);
  copyClusters(curr_best, cs, m);
  copyPrototypes(curr_best_protos, protos, m);
  //printResultsNoLC(curr_best_dist, phs, class, n);

  while (done < r)
    {
      swapPrototype(phs,cs,protos,n,m,&from,&to);
      curr_dist = kMedoids_Proposed(phs,cs,protos,n,m,plen,iters);
      if (curr_best_dist > curr_dist)
        {
          resetPhsAndClusters(curr_best, m);
	  copyPhsAndClusters(curr_best, cs, m);
	  resetPrototypes(curr_best_protos, m);
	  copyPrototypes(curr_best_protos, protos, m);
	  curr_best_dist = curr_dist;
	}
      else
	{
	  resetPhsAndClusters(cs, m);
	  copyPhsAndClusters(cs, curr_best, m);
	  resetPrototypes(protos, m);
	  copyPrototypes(protos, curr_best_protos, m);
	}
      done++;
    }

  resetPhsAndClusters(cs, m);
  copyPhsAndClusters(cs, curr_best, m);
  resetPrototypes(protos, m);
  copyPrototypes(protos, curr_best_protos, m);

  freeClusters(curr_best,m);
  freePrototypes(curr_best_protos,m);

  return curr_best_dist;
}

/* For deterministic RLS */
/*
void decideSwappableCluster(phoneme **phs, cluster **cs, int n, int m, int *f, int *t) 
{
  int i,j,k;
  int second_c = 0;
  double diff_sum = 0.0;
  double curr_2nd_DTW;
  double curr_2nd_min = DBL_MAX;
  double *differences = (double*)malloc(m*sizeof(double));

  for (i=0; i < m; i++, diff_sum = 0)                             // for each cluster calc matching
    {      
      for(j=0; j < cs[i]->phAmount; j++, curr_2nd_min = DBL_MAX)  // each phoneme in that cluster to
	{
	  for(k=0; k < m; k++)                                    // each other cluster
	    {
	      if (i != k)
		{
		  curr_2nd_DTW = DTW(cs[i]->phs[j], cs[k]->exemplar);
		  if ((curr_2nd_min > curr_2nd_DTW))
		    {
		      second_c = k;
		      curr_2nd_min = curr_2nd_DTW;
		    }
		}
	    }
	  diff_sum += curr_2nd_min - cs[i]->phs[j]->distortion;
	}
      differences[i] = diff_sum;
    }
  
  *f = smallestOf(differences,m);
  *t = biggestOf(differences,m);
  free(differences);
}
*/

double avgClusterDistance(cluster *cs1, cluster *cs2)
{
  int i, j;
  double sum;
  double distance;

  sum = 0.0;
  for (i = 0; i < cs1->phAmount; i++)
    {
      for (j = 0; j < cs2->phAmount; j++)
	{
	  sum += DTW(cs1->phs[i],cs2->phs[j]);
	}
    }
 
  //distance = sum * ((cs1->phAmount*cs2->phAmount)/((double)(cs1->phAmount + cs2->phAmount)));
  distance = sum / ((double)(cs1->phAmount + cs2->phAmount));

  return distance;
}

void initHC(phoneme **phs, cluster **cs, int n, int target_c)
{
  int i;

  if ((target_c < 1) || (target_c > n))
    {
      printf("Incorrect target amount for clusters\n");
      exit(-1);
    }
  
  for (i=0; i < n; i++)
    {
      phs[i]->distortion = 0.0;
      phs[i]->myCluster = i;
      cs[i]->phs[0] = phs[i];
      cs[i]->exemplar = phs[i];
      cs[i]->phAmount = 1;
      cs[i]->distortion = 0.0;
    }
}

double** calculateAvgClusterDistances(cluster **cs, int n)
{
  int i, j;
  double **dT;

  dT = (double**)malloc((n-1)*sizeof(double*));
  for (i=0; i < (n-1); i++)
    {
      dT[i] = (double*)malloc((n-i-1)*sizeof(double));
    }
  
  // calculate distances
  for (i=0; i < (n-1); i++)
    {
      for (j=0; j < (n-i-1); j++) 
	{
	  dT[i][j] = avgClusterDistance(cs[i], cs[n-j-1]);
	}  
    }

  return dT;
}

double HC(phoneme **phs, cluster **cs, int n, int target_c)
{
  int i, j, curr_cAmount;
  int receiver, donater;
  double curr_min;
  double **dT = NULL;

  dT = calculateAvgClusterDistances(cs,n);

  curr_cAmount = n;  

  while (curr_cAmount > target_c)
  {
    curr_min = DBL_MAX; 
    receiver = -1; 
    donater = -1;
    
    // pickLow; -------------V
    for (i=0; i < (n-1); i++)
      {
	for (j=0; j < (n-i-1); j++) 
	  {
	    if (curr_min > dT[i][j])
	      {
		curr_min = dT[i][j];
		receiver = i;
		donater = n - j - 1;
	      }
	  }
      }
    
    // merge; ------------V
    for (i=0; i < cs[donater]->phAmount; i++)
      {
	cs[donater]->phs[i]->myCluster = receiver;
	cs[receiver]->phs[cs[receiver]->phAmount + i] = cs[donater]->phs[i];
      	//cs[donater]->phs[i] = NULL;
      }
    cs[receiver]->phAmount += cs[donater]->phAmount;
    cs[donater]->phAmount = 0;
    cs[donater]->exemplar = NULL;

    // modifyDistance;  -----------V
    j = n - donater - 1;
    for(i=0; i < donater; i++)
      dT[i][j] = DBL_MAX;
    for(i=0; i < j; i++)
      dT[donater][i] = DBL_MAX;

    j = n - receiver - 1; 
    for(i=0; i < receiver; i++)
      if (cs[i]->exemplar != NULL)
	dT[i][j] = avgClusterDistance(cs[i], cs[receiver]);
    for(i=0; i < j; i++)
      if (cs[n-i-1]->exemplar != NULL)
	dT[receiver][i] = avgClusterDistance(cs[receiver], cs[n-i-1]);

    curr_cAmount--;
  }

  // crop
  i = 0; j = 0;
  while (j < target_c)
    {
      if (cs[j]->phAmount == 0)
	{
	  i = j + 1;
	  while (cs[i]->phAmount == 0)
	    i++;

	  copyPhAndCluster(cs[j],cs[i]);
	  resetCluster(cs[i]);
	  cs[j]->cNo = j;
	  for(receiver=0; receiver < cs[j]->phAmount; receiver++)
	    {
	      cs[j]->phs[receiver]->myCluster = j;
	    }
	}
      j++;
    }

  cs = (cluster**)realloc(cs, target_c*sizeof(cluster*));
  assignMedoids(cs,target_c);

  for (i=0; i < (n-1); i++) 
    free(dT[i]);
  free(dT);

  return calculateDistortion(n,cs,target_c);
}

double HC_Prototype(phoneme **phs, cluster **cs, phoneme **protos, int n, int target_c, int plen)
{
  int i, j, curr_cAmount;
  int receiver, donater;
  double curr_min;
  double **dT = NULL;

  dT = calculateAvgClusterDistances(cs,n);

  curr_cAmount = n;  

  while (curr_cAmount > target_c)
  {
    curr_min = DBL_MAX; 
    receiver = -1; 
    donater = -1;
    
    // pickLow; -------------V
    for (i=0; i < (n-1); i++)
      {
	for (j=0; j < (n-i-1); j++) 
	  {
	    if (curr_min > dT[i][j])
	      {
		curr_min = dT[i][j];
		receiver = i;
		donater = n - j - 1;
	      }
	  }
      }
    
    // merge; ------------V
    for (i=0; i < cs[donater]->phAmount; i++)
      {
	cs[donater]->phs[i]->myCluster = receiver;
	cs[receiver]->phs[cs[receiver]->phAmount + i] = cs[donater]->phs[i];
      	//cs[donater]->phs[i] = NULL;
      }
    cs[receiver]->phAmount += cs[donater]->phAmount;
    cs[donater]->phAmount = 0;
    cs[donater]->exemplar = NULL;

    // modifyDistance;  -----------V
    j = n - donater - 1;
    for(i=0; i < donater; i++)
      dT[i][j] = DBL_MAX;
    for(i=0; i < j; i++)
      dT[donater][i] = DBL_MAX;

    j = n - receiver - 1; 
    for(i=0; i < receiver; i++)
      if (cs[i]->exemplar != NULL)
	dT[i][j] = avgClusterDistance(cs[i], cs[receiver]);
    for(i=0; i < j; i++)
      if (cs[n-i-1]->exemplar != NULL)
	dT[receiver][i] = avgClusterDistance(cs[receiver], cs[n-i-1]);

    curr_cAmount--;
  }

  // crop
  i = 0; j = 0;
  while (j < target_c)
    {
      if (cs[j]->phAmount == 0)
	{
	  i = j + 1;
	  while (cs[i]->phAmount == 0)
	    i++;

	  //copyPrototype(protos[j],protos[i]); 
	  //resetPrototype(protos[i]);
	  copyPhAndCluster(cs[j],cs[i]);
	  resetCluster(cs[i]);
	  cs[j]->cNo = j;
	  for(receiver=0; receiver < cs[j]->phAmount; receiver++)
	    {
	      cs[j]->phs[receiver]->myCluster = j;
	    }
	}
      j++;
    }

  cs = (cluster**)realloc(cs, target_c*sizeof(cluster*));
  protos = (phoneme**)realloc(protos, target_c*sizeof(phoneme*));
  assignMedoids_Prototype(cs,target_c,plen);
  updatePrototypes(protos,cs,target_c);

  for (i=0; i < (n-1); i++) 
    free(dT[i]);
  free(dT);

  return calculateDistortion(n,cs,target_c);
}

double HC_Proposed(phoneme **phs, cluster **cs, phoneme **protos, int n, int target_c, int plen)
{
  int i, j, curr_cAmount;
  int receiver, donater;
  double curr_min, exProtoDist;
  double **dT = NULL;

  dT = calculateAvgClusterDistances(cs,n);

  curr_cAmount = n;  

  while (curr_cAmount > target_c)
  {
    curr_min = DBL_MAX; 
    receiver = -1; 
    donater = -1;
    
    // pickLow; -------------V
    for (i=0; i < (n-1); i++)
      {
	for (j=0; j < (n-i-1); j++) 
	  {
	    if (curr_min > dT[i][j])
	      {
		curr_min = dT[i][j];
		receiver = i;
		donater = n - j - 1;
	      }
	  }
      }
    
    // merge; ------------V
    for (i=0; i < cs[donater]->phAmount; i++)
      {
	cs[donater]->phs[i]->myCluster = receiver;
	cs[receiver]->phs[cs[receiver]->phAmount + i] = cs[donater]->phs[i];
      	//cs[donater]->phs[i] = NULL;
      }
    cs[receiver]->phAmount += cs[donater]->phAmount;
    cs[donater]->phAmount = 0;
    //protos[donater] = NULL;
    cs[donater]->exemplar = NULL;

    // modifyDistance;  -----------V
    j = n - donater - 1;
    for(i=0; i < donater; i++)
      dT[i][j] = DBL_MAX;
    for(i=0; i < j; i++)
      dT[donater][i] = DBL_MAX;

    j = n - receiver - 1; 
    for(i=0; i < receiver; i++)
      if (cs[i]->exemplar != NULL)
	dT[i][j] = avgClusterDistance(cs[i], cs[receiver]);
    for(i=0; i < j; i++)
      if (cs[n-i-1]->exemplar != NULL)
	dT[receiver][i] = avgClusterDistance(cs[receiver], cs[n-i-1]);

    curr_cAmount--;
  }

  // crop
  i = 0; j = 0;
  while (j < target_c)
    {
      if (cs[j]->phAmount == 0)
	{
	  i = j + 1;
	  while (cs[i]->phAmount == 0)
	    i++;

	  //copyPrototype(protos[j],protos[i]); // bug was here?
	  //resetPrototype(protos[i]);
	  copyPhAndCluster(cs[j],cs[i]);
	  resetCluster(cs[i]);
	  cs[j]->cNo = j;
	  for(receiver=0; receiver < cs[j]->phAmount; receiver++)
	    {
	      cs[j]->phs[receiver]->myCluster = j;
	    }
	  //updatePrototype(protos[j],cs[j]);
	}
      j++;
    }

  cs = (cluster**)realloc(cs, target_c*sizeof(cluster*));
  protos = (phoneme**)realloc(protos, target_c*sizeof(phoneme*));
  assignMedoids_Prototype(cs,target_c,plen);
  updatePrototypes(protos,cs,target_c);
  for (i=0; i < target_c; i++)
    {
      exProtoDist = DBL_MAX;
      calculateInternalDistortion(cs[i]);
      
      do
	{
	  assignPathsForCluster(cs[i]->phs, cs[i]->exemplar, cs[i]->phAmount);
	  updatePrototype(protos[i], cs[i]);
	  
	  exProtoDist = cs[i]->distortion;
	  calculateInternalDistortion(cs[i]);
	}
      while (RelDif(cs[i]->distortion, exProtoDist) > TOLERANCE);
    }

  for (i=0; i < (n-1); i++) 
    free(dT[i]);
  free(dT);

  return calculateDistortion(n,cs,target_c);
}

int main(int argc, char **argv)
{
  /* declarations */
  int rls_half = 1;
  int rls_count = 2;
  int counter = 1;
  int km = 0;
  int chdir_ok = -1;
  int n_size = 0;
  int path_size = 0;
  int i;
  double totalDistortion = 0.0;
  double exProtoDist = DBL_MAX;
  parameters *params;

  cluster **cs = NULL;
  cluster **original = NULL;
  phoneme **phs = NULL;
  phoneme **protos = NULL;
  phoneme **protos2 = NULL;
  phoneme **classification = NULL;

  /* initialize */
  srand((unsigned)time(NULL));
  
  if (argc <= 1)
    {
      printUsage();
      exit(1);
    }

  params = initParams();
  processParams(params, argc, argv);

  if (params->oneFile == 0)
    {
      chdir_ok = chdir(params->indir);

      if (chdir_ok != 0)
	{
	  printf("Error while trying to move to directory \"%s\"\n", params->indir);
	  freeParams(params);
	  exit(2);
	}
      phs = readAllFiles(params->indir, &n_size, &path_size);
    }
  else
    phs = readDatasetFile(params->indir, &n_size);

  /* Check k < n */
  if (params->clusters > n_size)
    {
      printf("Given cluster amount exceeds amount of phonemes\n");
      exit(3);
    }

  /*
  if ((n_size == 600) || (n_size == 640) || (n_size == 925) || (n_size == 2565))
    {
      classification = newPhonemes(n_size);
      for (i=0; i < n_size; i++) {
	classification[i] = newPhoneme(phs[i]->filename,phs[i]->dimensions);
      }
      createClassification(classification,n_size);
    }
  */

  /* calculate stuff */
  switch (params->cmethod) {
  case 0:
    cs = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    totalDistortion = kMedoids(phs, cs, n_size, params->clusters, &(params->kmIters));
    break;
  case 1:
    cs = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    totalDistortion = iter_kMedoids(phs, cs, n_size, params->clusters, params->iterations, &(params->kmIters));
    break;
  case 2:
    cs = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);  
    totalDistortion = RLS(phs, cs, n_size, params->clusters, params->iterations, &(params->kmIters));
    break;
  case 3:
    cs = newClusters(n_size, n_size);
    initHC(phs, cs, n_size, params->clusters);
    totalDistortion = HC(phs, cs, n_size, params->clusters);
    break;
  case 4:
    cs = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, params->clusters);
    totalDistortion = kMedoids_Prototype(phs, cs, protos, n_size, params->clusters, path_size, &(params->kmIters));
    break;
  case 5:
    cs = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);  
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, params->clusters);
    totalDistortion = RLS_Prototype(phs, cs, protos, n_size, params->clusters, params->iterations, path_size, &(params->kmIters));
    break;
  case 6:
    cs = newClusters(n_size, n_size);
    initHC(phs, cs, n_size, params->clusters);
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, n_size);
    totalDistortion = HC_Prototype(phs, cs, protos, n_size, params->clusters, path_size);
    break;
  case 7:
    cs = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, params->clusters);
    totalDistortion = kMedoids_Proposed(phs, cs, protos, n_size, params->clusters, path_size, &(params->kmIters));
    break;
  case 8:
    cs = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, params->clusters);
    totalDistortion = RLS_Proposed(phs, cs, protos, n_size, params->clusters, params->iterations, path_size, &(params->kmIters));
    break;
  case 9:
    cs = newClusters(n_size, n_size);
    initHC(phs, cs, n_size, params->clusters);
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, n_size);
    totalDistortion = HC_Proposed(phs, cs, protos, n_size, params->clusters, path_size);
    break;
    /*
  case 10:
    cs = newClusters(n_size, params->clusters);
    original = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    copyClusters(original, cs, params->clusters);

    printf("%d\t",params->clusters);
    totalDistortion = RLS_Class(phs, cs, classification, n_size, params->clusters, params->iterations, &(params->kmIters));
    printResultsNoLC(totalDistortion, phs, classification, n_size);
    resetPhsAndClusters(cs, params->clusters);
    copyClusters(cs, original, params->clusters);

    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, params->clusters);
    protos2 = newPrototypes(cs, params->clusters);

    totalDistortion = RLS_Prototype_Class(phs, cs, protos, classification, n_size, params->clusters, params->iterations, path_size, &(params->kmIters));
    printResultsNoLC(totalDistortion, phs, classification, n_size);
    resetPhsAndClusters(cs, params->clusters);
    copyClusters(cs, original, params->clusters);

    totalDistortion = RLS_Proposed_Class(phs, cs, protos2, classification, n_size, params->clusters, params->iterations, path_size, &(params->kmIters));
    printResultsNoLC(totalDistortion, phs, classification, n_size);
    printf("\n");
    break;
    */
  case 11:
    cs = newClusters(n_size, n_size);
    initHC(phs, cs, n_size, params->clusters);
    totalDistortion = HC(phs, cs, n_size, params->clusters);
    printResultsNoLC(totalDistortion,phs,classification,n_size);
    totalDistortion = kMedoids(phs, cs, n_size, params->clusters, &(params->kmIters));    
    printResultsNoLC(totalDistortion,phs,classification,n_size);
    //    printf("%f\t",params->clusters,totalDistortion);
    break;
  case 12:
    cs = newClusters(n_size, n_size);
    initHC(phs, cs, n_size, params->clusters);
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, n_size);
    totalDistortion = HC_Prototype(phs, cs, protos, n_size, params->clusters, path_size);
    printResultsNoLC(totalDistortion,phs,classification,n_size);
    totalDistortion = kMedoids_Prototype(phs, cs, protos, n_size, params->clusters, path_size, &(params->kmIters));
    printResultsNoLC(totalDistortion,phs,classification,n_size);
    break;
  case 13:
    cs = newClusters(n_size, n_size);
    initHC(phs, cs, n_size, params->clusters);
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, n_size);
    totalDistortion = HC_Proposed(phs, cs, protos, n_size, params->clusters, path_size);
    printResultsNoLC(totalDistortion,phs,classification,n_size);
    totalDistortion = kMedoids_Proposed(phs, cs, protos, n_size, params->clusters, path_size, &(params->kmIters));
    printResultsNoLC(totalDistortion,phs,classification,n_size);
    //    printf("\n");
    break;
  case 14: //iters
    cs = newClusters(n_size, params->clusters);
    original = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    copyClusters(original, cs, params->clusters);
    
    // printf("-Medoid-\n");
    totalDistortion = kMedoids(phs,cs,n_size,params->clusters,&(params->kmIters));
    printf("%f\t", totalDistortion);

    resetPhsAndClusters(cs, params->clusters);
    copyClusters(cs, original, params->clusters);

    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, params->clusters);
    protos2 = newPrototypes(cs, params->clusters);

    // printf("-Prototype-\n");
    totalDistortion = kMedoids_Prototype(phs,cs,protos,n_size,params->clusters,path_size,&(params->kmIters));
    resetPhsAndClusters(cs, params->clusters);
    copyClusters(cs, original, params->clusters);
    printf("%f\t", totalDistortion);

    //printf("-Proposed-\n");
    totalDistortion = kMedoids_Proposed(phs,cs,protos,n_size,params->clusters,path_size,&(params->kmIters));
    printf("%f\t", totalDistortion);

    //printf("\n");
    break;
  case 15:
    cs = newClusters(n_size, n_size);
    original = newClusters(n_size, params->clusters);
    initHC(phs, cs, n_size, params->clusters);

    totalDistortion = HC(phs, cs, n_size, params->clusters);
    copyClusters(original, cs, params->clusters);
    printf("%f\t", totalDistortion);
    totalDistortion = kMedoids(phs, cs, n_size, params->clusters, &(params->kmIters));    
    printf("%f\t", totalDistortion);

    resetPhsAndClusters(cs, params->clusters);
    copyPhsAndClusters(cs, original, params->clusters);

    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, params->clusters);
    assignMedoids_Prototype(cs,params->clusters,path_size);
    updatePrototypes(protos,cs,params->clusters);
    printf("%f\t", calculateDistortion(n_size,cs,params->clusters));
    totalDistortion = kMedoids_Prototype(phs,cs,protos,n_size,params->clusters,path_size,&(params->kmIters));
    printf("%f\t", totalDistortion);
    resetPhsAndClusters(cs, params->clusters);
    copyPhsAndClusters(cs, original, params->clusters);

    protos2 = newPrototypes(cs, params->clusters);
    assignMedoids_Prototype(cs, params->clusters, path_size);
    updatePrototypes(protos2,cs,params->clusters);
    for (i=0; i < params->clusters; i++)
      {
	exProtoDist = DBL_MAX;
	calculateInternalDistortion(cs[i]);
	
	do
	  {
	    assignPathsForCluster(cs[i]->phs, cs[i]->exemplar, cs[i]->phAmount);
	    updatePrototype(protos2[i], cs[i]);
	    
	    exProtoDist = cs[i]->distortion;
	    calculateInternalDistortion(cs[i]);
	  }
	while (RelDif(cs[i]->distortion, exProtoDist) > TOLERANCE);
      }    
    printf("%f\t", calculateDistortion(n_size,cs,params->clusters));
    totalDistortion = kMedoids_Proposed(phs,cs,protos2,n_size,params->clusters,path_size,&(params->kmIters));
    printf("%f\t", totalDistortion);
    //    printf("\n");
    break;
  case 16:
    cs = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, params->clusters);
    totalDistortion = iter_kMedoids_Prototype(phs, cs, protos, n_size, params->clusters, params->iterations, path_size, &(params->kmIters));
    break;
  case 17:
    cs = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    allocateDTWpaths(phs, n_size, path_size);
    protos = newPrototypes(cs, params->clusters);
    totalDistortion = iter_kMedoids_Proposed(phs, cs, protos, n_size, params->clusters, params->iterations, path_size,&(params->kmIters));
    break;
    /*
  case 20:
    //allocateDTWpaths(phs, n_size, path_size);
    printf("%s\n",phs[0]->filename);
    for (i=1; i < 10; i++)
      {
	totalDistortion = optimalSegmentationDistance(phs[i],phs[0]);
	printf("GPS%d(%d): %.2f\n", i, phs[i]->frameAmount, totalDistortion);
      }
    break;
  case 30: // RLS-tests
    cs = newClusters(n_size, params->clusters);
    original = newClusters(n_size, params->clusters);
    getRandomCodebook(phs, cs, n_size, params->clusters);
    copyClusters(original, cs, params->clusters);

    rls_count = 256;
    for (counter=1; counter < 11; counter++) 
      {
	km = counter;

	totalDistortion = RLS(phs, cs, n_size, params->clusters, rls_count, &km);
	resetPhsAndClusters(cs, params->clusters);
	copyClusters(cs, original, params->clusters);
	printf("\n");
      }
    break;
    */
  default:
    printf("Unknown clustering method!\n");
    printUsage();
    exit(4);
  }

  /* print stuff */
  //printPhonemes(phs);
  //printf("\n");
  //printf("*** FINAL RESULT FOR %d PHONEMES ***\n", n_size);
  //printClusterStatus(cs, params->clusters);
  //printClusterStatusSimple(cs, params->clusters);

  if (params->cmethod < 10)
    {
        printf("%f\t", totalDistortion);
	/*  	if (classification != NULL) 
	  printf("%f", ClassificationError(phs, classification, n_size));
	  printf("\n");*/
    }
  
  
  /* free stuff */
  if (cs != NULL)
    freeClusters(cs,params->clusters);
  freePhonemes(phs, n_size);
  if (protos != NULL)
    freePrototypes(protos,params->clusters);
  if (classification != NULL)
    freePhonemes(classification, n_size);
  freeParams(params);

  return 0;
}
